#include <iostream>
#include <strstream>
#include <fstream>
#include <string>
#include <stack>
#include <cmath>
#include <cstdio>
#include <cassert>
#include <cstdlib>
#include <cstring>
using namespace std;

#include <GL/glut.h>		// access to glut and OpenGL libraries

// The width and height of our window.  I choose values that
// ensure that the display has the same proportions that
// a sheet of paper has (i.e., 8.5" x 11.0").

const double PageW = 8.5;
const double PageH = 11.0;
const double AspectRatioOfPage = PageW / PageH;

const int H = 700;					// Height of window, in pixels
const int W = (int) (H * AspectRatioOfPage);// Width of window, in pixels

// The default units in Postscript are "points".
// A "point" is 1/72".  We use this to determine the
// width and height of a page, in points.

const double  PtsPerInch = 72.0;
const double  PageWidthPts = PtsPerInch * PageW;
const double  PageHeightPts = PtsPerInch * PageH;

// Here are the string constants corresponding to the
// Postscript commands we are processing.

const string MOVEREL = "rmoveto";
const string MOVETO = "moveto";
const string LINEREL = "rlineto";
const string LINETO = "lineto";
const string ADD = "add";
const string SUB = "sub";
const string MUL = "mul";
const string DIV = "div";
const string DUP = "dup";
const string COPY = "copy";
const string ARC = "arc";
const string STROKE = "stroke";

typedef stack<double> OperandStack;

OperandStack stk;
string fname;		// Keep track of name of input file
double curX;
double curY;

float zoom = 1;
int xpan = PageWidthPts * zoom / 2;
int ypan = PageHeightPts * zoom / 2;

// d2r
// This function converts from degrees to radians.
double d2r(double deg)
{ 
	const double PI = 3.14159265358979323846;
	return PI * deg / 180.0;
}

// Pop
// This function performs the dual function of popping the
// stack, and returning the value of the popped element.
double Pop(OperandStack &stk)
{
	assert(!stk.empty());
	double val = stk.top();
	stk.pop();
	return val;
}

// dumpStack
void dumpStack(stack<double> stk)
{
	cout << "[";
	while (!stk.empty()) {
		cout << stk.top() << ' ';
		stk.pop();
	}
	cout << "]" << endl;
}

// GetCircularCoordinates
// This function finds the (x,y) of a point on a circle, where
// the radius, center point, and theta are given.
//  (cx, cy) - center point of circle
//  r - radius of circle
//  ang - angle of point to be evaluated (in degrees)
//  (x, y) - these values are determined by this function and
//		passed out to the calling routine

void GetRadialCoordinates(double cx, double cy, double r, double ang, double &x, double &y)
{
	double theta = d2r(ang);
	x = r * cos(theta) + cx;
	y = r * sin(theta) + cy;
}

double popTop(){
	return Pop(stk);
}

void popStack(double &val){
	val = popTop();
}

void popStack(double &x, double &y){
	y = popTop();
	x = popTop();
}

void lineTo(int x, int y){
	glBegin(GL_LINES);
		glVertex2d(curX, curY);
		glVertex2d(x, y);
	glEnd();	
	curX = x;
	curY = y;
}

void rLineTo(int x, int y){
	lineTo(curX + x, curY + y);
}

void incCircle(double &theta){
	if(++theta > 360)
		theta = 0;
}

// DrawArc
void DrawArc(double cx, double cy, double r, double ang1, double ang2)
{
	double x, y, nextX, nextY;

	GetRadialCoordinates(cx, cy, r, ang1, x, y);

	if((curX != -1 && curY != -1) && (x != curX || y != curY))
		lineTo(x, y);

	// Create arc by drawing a bunch of little line segments.
	for (double theta=ang1; theta != ang2; incCircle(theta)) {
		GetRadialCoordinates(cx, cy, r, theta, nextX, nextY);
		glBegin(GL_LINES);
			glVertex2d(x, y);
			glVertex2d(nextX, nextY);
		glEnd();
		x = nextX;
		y = nextY;
	}

	GetRadialCoordinates(cx, cy, r, ang2, curX, curY);
	glBegin(GL_LINES);
		glVertex2d(x, y);
		glVertex2d(curX, curY);
	glEnd();
}

// ProcessToken
// This function processes a single token in the PS file.
// The token is either a number or a operator.  If it is
// a number, it is pushed on the stack.  If it is a operator,
// the operation is performed.

void ProcessToken(const string &item)
{	
	double val;

	// sscanf is a routine that reads from a string.
	// The "%lf" indicates that it should try to read
	// a "double"  from the string.
	// If successful, sscanf_s will return 1 and set
	// val to the value of the number.

	if (sscanf(item.c_str(), "%lf", &val)) { //was sscanf_s changed to work in linux
		stk.push(val);
	} else if (item == MOVEREL) {
		double x,y;
		popStack(x, y);
		curX += x;
		curY += y;
	} else if (item == DUP) {
		double x = popTop();
		stk.push(x);
		stk.push(x);
	} else if (item == COPY) {
		double n = popTop();
		stack<double> tempstk;
		for(int x = 0; x < n; x++)
			tempstk.push(popTop());
		stack<double> tempstk2 = tempstk;//make 2 copies
		for(int x = 0; x < n; x++)
			stk.push(Pop(tempstk));
		for(int x = 0; x < n; x++)
			stk.push(Pop(tempstk2));
	} else if (item == ADD) {		
		stk.push(popTop() + popTop());
	} else if (item == MUL) {
		stk.push(popTop() * popTop());
	} else if (item == SUB) {
		double num2 = popTop();
		double num1 = popTop();
		stk.push(num1 - num2);
	} else if (item == DIV) {
		double num2 = popTop();
		double num1 = popTop();
		stk.push(num1 / num2);
	} else if (item == MOVETO) {
		popStack(curX, curY);
	} else if (item == LINEREL) {
		double x,y;
		popStack(x, y);
		rLineTo(x, y);
	} else if (item == LINETO) {
		double x,y;
		popStack(x, y);
		lineTo(x, y);
	} else if (item == ARC) {
		double cx, cy, r, ang1, ang2;
		popStack(ang2);
		popStack(ang1);
		popStack(r);
		popStack(cy);
		popStack(cx);		
		DrawArc(cx, cy, r, ang1, ang2);
	} else if (item == STROKE) {
		;//ignore this, already drawing black line
	} else {
		cout << "Operator: " << item << " not supported" << endl;
	}
}

// ProcessLine
// This function processes a single line of the PS file.  The
// line may be a comment line or a line containing many tokens.
// If the line is a comment line, it is disregared.  Otherwise,
// the individual tokens of the line are processed.

void ProcessLine(const string &line)
{
	istrstream iss(line.c_str());	// use a string like "cin"

	while (!iss.eof()) {
		string item;
		iss >> item;						// read next token (i.e., word)
		if (item == "" || item[0] == '%')	// stop when no more tokens on line
											// or a comment is seen
			break;

		ProcessToken(item);
	}
}

// display
// This callback is invoked automatically by the window system.
// It is called as a request to "redraw the picture."  Events
// that cause this function to be called include:
//
//			Window is covered by a window and then uncovered
//			Window is iconified and then opened
//
// This function processes a complete PS file.  This is done
// by opening the file and reading individual lines.  Each line
// is processed individually.
//
// We need to reprocess the file each time this function
// is called.  Recall, the state variable contains the
// name of the file.  The process of continually
// reopening the file is inefficient; we are just 
// starting to learn....


void display(void)
{
	curX = -1;
	curY = -1;
	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT);	// Blank the window

	glColor3f(0,0,0);

	stk = stack<double>();

	ifstream is(fname.c_str());

	if (is.good()) {
		while (!is.eof()) {
			string line;
			getline(is, line);
			ProcessLine(line);
		}
	} else {
		cout << "Can't find file: " << fname << endl;
		exit(0);
	}

	is.close();

	glutSwapBuffers();
}

// This is code needed to initialize the window.  The purpose
// is to ensure that the window has the proportions of a
// page of paper.  We will discuss these functions in the
// coming weeks.

void init(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, PageWidthPts, 0, PageHeightPts);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void reinit()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(xpan - PageWidthPts * zoom / 2, xpan + PageWidthPts * zoom / 2, ypan - PageHeightPts * zoom / 2, ypan + PageHeightPts * zoom / 2);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void keyboard(unsigned char key, int x, int y)
{
	const float step = 1.25;
	switch (key) {
		case '-':	zoom *= step;
				break;
		case '+':	zoom /= step;
				break;
	}
	cout << zoom << endl;
	reinit();
	glutPostRedisplay();
}

void special(int key, int x, int y)
{
	const int step = 10;
	switch (key) {
		case GLUT_KEY_LEFT:	xpan += step;
					break;
		case GLUT_KEY_RIGHT:	xpan -= step;
					break;
		case GLUT_KEY_DOWN:	ypan += step;
					break;
		case GLUT_KEY_UP:	ypan -= step;
					break;
	}
	reinit();
	glutPostRedisplay();
}

int main(int argc, char *argv[])
{
	cout << "Filename: ";
	cin >> fname;

	// Initialize the glut library
	glutInit(&argc, argv);

	// glutInitDisplayMode() is used to inform the glut
	// library about the type of properties we want our
	// display to have.
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);	

	// glutInitWindowPosition() used to inform the glut
	// library about where we want the window to be
	// placed, relative to the entire screen.  Upper left
	// corner is (0, 0).
	glutInitWindowPosition(100, 100);

	glutInitWindowSize(W, H);
	glutCreateWindow("Postscript viewer by Travis Burtrum");

	// The window system will call my function, display(),
	// whenever my display needs to be updated (e.g., window
	// is covered and then exposed, window is iconified and
	// then opened.  display() is referred to as a "callback"
	// and glutDisplayFunc() is used for "callback registration".

	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(special);
	
	init(W, H);

	// glutMainLoop() is the main event handler.  This glut
	// routine simply looks for things to happen (e.g.,
	// mouse click, key pressed, etc.)  When such an event
	// occurs, your appropriate event handler is called.  In
	// this program, we have only one event handler defined.
	// Thus, events such as mouse clicks and keys pressed
	// will be ignored.

	glutMainLoop();
}
